from mypy_extensions import u8

class ReadBuffer:
    def __init__(self, source: bytes) -> None: ...

class WriteBuffer:
    def getvalue(self) -> bytes: ...

def write_bool(data: WriteBuffer, value: bool, /) -> None: ...
def read_bool(data: ReadBuffer, /) -> bool: ...
def write_str(data: WriteBuffer, value: str, /) -> None: ...
def read_str(data: ReadBuffer, /) -> str: ...
def write_bytes(data: WriteBuffer, value: bytes, /) -> None: ...
def read_bytes(data: ReadBuffer, /) -> bytes: ...
def write_float(data: WriteBuffer, value: float, /) -> None: ...
def read_float(data: ReadBuffer, /) -> float: ...
def write_int(data: WriteBuffer, value: int, /) -> None: ...
def read_int(data: ReadBuffer, /) -> int: ...
def write_tag(data: WriteBuffer, value: u8, /) -> None: ...
def read_tag(data: ReadBuffer, /) -> u8: ...
def cache_version() -> u8: ...
def extract_symbol(data: ReadBuffer, /) -> bytes: ...
