#ifndef LIBRT_THREADING_H
#define LIBRT_THREADING_H

#include <Python.h>

#define LIBRT_THREADING_ABI_VERSION 1
#define LIBRT_THREADING_API_VERSION 1
#define LIBRT_THREADING_API_LEN 8

#endif  // LIBRT_THREADING_H
