#ifndef VECS_INTERNAL_H_INCL
#define VECS_INTERNAL_H_INCL

// Internal header for mypyc/lib-rt/vecs implementation

#include <Python.h>

// The base vec type
extern PyTypeObject VecType;

#endif  // VECS_INTERNAL_H_INCL
